# RAD-tools - program for spin Hamiltonian and magnons.
# Copyright (C) 2022-2024  Andrey Rybakov
#
# e-mail: anry@uv.es, web: adrybakov.com
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from os.path import abspath, isfile, join

import matplotlib.pyplot as plt
import numpy as np
from termcolor import cprint

from radtools.decorate.axes import plot_hlines, plot_vlines
from radtools.decorate.colormap import custom_cmap
from radtools.dos.dos import prepare_custom_pdos
from radtools.dos.pdos import PDOS
from radtools.dos.plotting import COLOURS

__all__ = [
    "plot_custom_fatbands",
    "plot_fatbands",
]


def plot_custom_fatbands(
    dos,
    custom,
    output_root=".",
    energy_window=None,
    k_window=None,
    efermi=0.0,
    interactive=False,
    separate=False,
    save_pickle=False,
    save_txt=False,
    colours=COLOURS,
    labels=None,
    legend_fontsize=12,
    axes_labels_fontsize=14,
    title_fontsize=18,
    k_points=None,
):
    cprint("Plotting custom plot", "green")
    print("Input is understood as:")
    projectors = []

    # Check if labels are provided and correct
    if labels is not None and len(labels) != len(custom):
        raise ValueError(
            f"Got {len(labels)} labels, but {len(custom)} PDOS, have to be the same."
        )

    # Set projectors
    if labels is None:
        projectors = custom
    else:
        projectors = labels

    # Get PDOS array
    pdos = prepare_custom_pdos(dos=dos, custom=custom, quiet=False)

    # Create PDOS object
    pdos = PDOS(
        energy=dos.energy,
        pdos=pdos,
        projectors_group="Total (sum)",
        projectors=projectors,
        spin_pol=dos.case in [2, 3],
    )

    # Compute output name
    if isfile(join(output_root, "custom.png")):
        i = 1
        while isfile(join(output_root, f"custom{i}.png")):
            i += 1
        output_name = f"custom{i}"
    else:
        output_name = "custom"

    # Save txt
    if save_txt:
        pdos.dump_txt(join(output_root, f"{output_name}.txt"))

    # Plot
    plot_fatbands(
        pdos=pdos,
        efermi=efermi,
        output_name=join(output_root, output_name),
        ylim=energy_window,
        xlim=k_window,
        interactive=interactive,
        separate=separate,
        save_pickle=save_pickle,
        colours=colours,
        legend_fontsize=legend_fontsize,
        axes_labels_fontsize=axes_labels_fontsize,
        title_fontsize=title_fontsize,
        k_points=k_points,
    )
    cprint(f"Result is in {abspath(join(output_root, f'{output_name}.png'))}", "blue")


def plot_fatbands(
    pdos: PDOS,
    efermi=0.0,
    output_name="fatbands",
    ylim=None,
    xlim=None,
    interactive=False,
    separate=False,
    save_pickle=False,
    colours=COLOURS,
    legend_fontsize=14,
    axes_labels_fontsize=12,
    title_fontsize=18,
    k_points=None,
):
    r"""
    Plot fatbands.

    Parameters
    ----------
    pdos : :py:class:`.PDOS`
        PDOS for the plot.
    efermi : float, default 0
        Fermi energy.
    output_name : str, default "pdos"
        output_name for the plot file. Extension ".png" is added at the end.
    ylim : tuple
        limits for the y (PDOS) axis
    interactive : bool, default False
        Whether to use interactive plotting mode.
    separate : bool, default False
        Whether to plot each entry in an individual figure.
    save_pickle : bool, default False
        Whether to save figure as a .pickle file.
        Helps for custom modification of particular figures.
    colours : list
        List of colours to be used. values are passed directly to matplotlib
    total_label : str or ``None``, default "default"
        Label for the total data. If None , then the label is not added
    axes_label_fontsize : int, default 14
        Fontsize of the axes labels.
    legend_fontsize : int, default 12
        Fontsize of the legend.
    title_fontsize : int, default 18
        Fontsize of the title
    """

    norm = np.amax(pdos.pdos)

    if xlim is None:
        xlim = (pdos.kpoints[0], pdos.kpoints[-1])
    if ylim is None:
        ylim = (pdos.energy[0], pdos.energy[-1])

    if k_points is not None:
        tmp = k_points
        k_points = [[], []]
        for i in range(len(tmp) // 2):
            k_points[1].append(tmp[2 * i])
            k_points[0].append(float(tmp[2 * i + 1]))

    if not separate:
        pass

    def plot_entry(axs, pdos, projector, xlim, ylim, efermi):
        if pdos.spin_pol:
            axs[1].get_yaxis().set_visible(False)
            if k_points is not None:
                axs[1].set_xticks(
                    k_points[0], k_points[1], fontsize=axes_labels_fontsize
                )
                plot_vlines(axs[1], k_points[0])
            else:
                axs[1].set_xlabel("k path", fontsize=axes_labels_fontsize)
            axs[1].set_ylim(*tuple(ylim))
            axs[1].set_xlim(*tuple(xlim))
        if k_points is not None:
            axs[0].set_xticks(k_points[0], k_points[1], fontsize=axes_labels_fontsize)
            plot_vlines(axs[0], k_points[0])
        else:
            axs[0].set_xlabel("k path", fontsize=axes_labels_fontsize)
        axs[0].set_ylim(*tuple(ylim))
        axs[0].set_xlim(*tuple(xlim))

        if efermi == 0:
            axs[0].set_ylabel("E, eV", fontsize=axes_labels_fontsize)
        else:
            axs[0].set_ylabel("E - E$_{Fermi}$, eV", fontsize=axes_labels_fontsize)

        if pdos.spin_pol:
            axs[0].imshow(
                pdos[projector][0].T,
                cmap=custom_cmap((1, 1, 1), (0, 0, 1)),
                origin="lower",
                extent=(
                    xlim[0],
                    xlim[1],
                    ylim[0],
                    ylim[1],
                ),
                aspect="auto",
                vmax=norm,
                vmin=0,
            )
            axs[0].set_title(f"{projector} (up)", fontsize=title_fontsize)
            axs[1].imshow(
                pdos[projector][1].T,
                cmap=custom_cmap((1, 1, 1), (1, 0, 0)),
                origin="lower",
                extent=(
                    xlim[0],
                    xlim[1],
                    ylim[0],
                    ylim[1],
                ),
                aspect="auto",
                vmax=norm,
                vmin=0,
            )
            axs[1].set_title(f"{projector} (down)", fontsize=title_fontsize)
        else:
            axs[0].imshow(
                pdos[projector].T,
                cmap="inferno",
                origin="lower",
                extent=(
                    xlim[0],
                    xlim[1],
                    ylim[0],
                    ylim[1],
                ),
                aspect="auto",
                vmax=norm,
                vmin=0,
            )
            axs[0].set_title(projector, fontsize=title_fontsize)

        if efermi != 0:
            plot_hlines(axs[0], 0)
            if pdos.spin_pol:
                plot_hlines(axs[1], 0)

    for i, projector in enumerate(pdos):
        if separate:
            if pdos.spin_pol:
                fig, axs = plt.subplots(1, 2, figsize=(8, 7))
            else:
                fig, axs = plt.subplots(figsize=(4, 7))
                axs = [axs]
            fig.subplots_adjust(wspace=0)
            plot_entry(axs, pdos, projector, xlim, ylim, efermi)
            if interactive:
                plt.show()
            else:
                plt.savefig(f"{output_name}_{i}.png", dpi=600, bbox_inches="tight")
                if save_pickle:
                    import pickle

                    with open(f"{output_name}_{i}.png.pickle", "wb") as file:
                        pickle.dump(fig, file)
            plt.close()
        else:
            plot_entry(axs[i], pdos, projector, xlim, ylim, efermi)

    if not separate:
        if interactive:
            plt.show()
        else:
            plt.savefig(f"{output_name}.png", dpi=600, bbox_inches="tight")
            if save_pickle:
                import pickle

                with open(f"{output_name}.png.pickle", "wb") as file:
                    pickle.dump(fig, file)
        plt.close()



# RAD-tools - program for spin Hamiltonian and magnons.
# Copyright (C) 2022-2024  Andrey Rybakov
#
# e-mail: anry@uv.es, web: adrybakov.com
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from math import cos, sqrt

import numpy as np
from scipy.spatial.transform import Rotation

from radtools.constants import TODEGREES, TORADIANS
from radtools.crystal.constants import ABS_TOL, ABS_TOL_ANGLE
from radtools.numerical import compare_numerically

__all__ = [
    "volume",
    "angle",
    "parallelepiped_check",
    "absolute_to_relative",
    "span_orthonormal_set",
]


def volume(*args):
    r"""
    Computes volume.

    .. versionadded:: 0.7

    Three type of arguments are expected:

    * One argument.
        Matrix ``cell``.
        Volume is computed as:

        .. math::
            V = v_1 \cdot (v_2 \times v_3)
    * Three arguments.
        Vectors ``v1``, ``v2``, ``v3``.
        Volume is computed as:

        .. math::
            V = v_1 \cdot (v_2 \times v_3)
    * Six arguments.
        Parameters ``a``, ``b``, ``c``, ``alpha``, ``beta``, ``gamma``.
        Volume is computed as:

        .. math::
            V = abc\sqrt{1+2\cos(\alpha)\cos(\beta)\cos(\gamma)-\cos^2(\alpha)-\cos^2(\beta)-\cos^2(\gamma)}

    Notes
    -----

    Volume can be negative if the three vectors are not right-handed.

    Parameters
    ----------
    v1 : (3,) |array_like|_
        First vector.
    v2 : (3,) |array_like|_
        Second vector.
    v3 : (3,) |array_like|_
        Third vector.
    cell : (3,3) |array_like|_
        Cell matrix, rows are interpreted as vectors.
    a : float, default 1
        Length of the :math:`v_1` vector.
    b : float, default 1
        Length of the :math:`v_2` vector.
    c : float, default 1
        Length of the :math:`v_3` vector.
    alpha : float, default 90
        Angle between vectors :math:`v_2` and :math:`v_3`. In degrees.
    beta : float, default 90
        Angle between vectors :math:`v_1` and :math:`v_3`. In degrees.
    gamma : float, default 90
        Angle between vectors :math:`v_1` and :math:`v_2`. In degrees.

    Returns
    -------
    volume : float
        Volume of corresponding region in space.
    """

    if len(args) == 1:
        cell = np.array(args[0])
    elif len(args) == 3:
        cell = np.array(args)
    elif len(args) == 6:
        a, b, c, alpha, beta, gamma = args
        alpha = alpha * TORADIANS
        beta = beta * TORADIANS
        gamma = gamma * TORADIANS
        sq_root = (
            1
            + 2 * cos(alpha) * cos(beta) * cos(gamma)
            - cos(alpha) ** 2
            - cos(beta) ** 2
            - cos(gamma) ** 2
        )
        return a * b * c * sqrt(sq_root)
    else:
        raise ValueError(
            "Unable to identify input. "
            + "Supported: one (3,3) array_like, or three (3,) array_like, or 6 floats."
        )

    return np.linalg.det(cell)


def angle(v1, v2, radians=False):
    r"""
    Angle between two vectors.

    .. versionadded:: 0.7

    .. math::

        \alpha = \dfrac{(\vec{v_1} \cdot \vec{v_2})}{\vert\vec{v_1}\vert\cdot\vert\vec{v_2}\vert}

    Parameters
    ----------
    v1 : (3,) |array_like|_
        First vector.
    v2 : (3,) |array_like|_
        Second vector.
    radians : bool, default False
        Whether to return value in radians. Return value in degrees by default.

    Returns
    -------
    angle: float
        Angle in degrees or radians (see ``radians``).

    Raises
    ------
    ValueError
        If one of the vectors is zero vector (or both). Norm is compared against
        :numpy:`finfo`\ (float).eps.
    """

    # Normalize vectors
    v1_norm = np.linalg.norm(v1)
    v2_norm = np.linalg.norm(v2)
    if abs(v1_norm) <= np.finfo(float).eps or abs(v2_norm) <= np.finfo(float).eps:
        raise ValueError("Angle is ill defined (zero vector).")

    v1 = np.array(v1) / v1_norm
    v2 = np.array(v2) / v2_norm

    alpha = np.arccos(np.clip(np.dot(v1, v2), -1.0, 1.0))
    if radians:
        return alpha
    return alpha * TODEGREES


def parallelepiped_check(a, b, c, alpha, beta, gamma, raise_error=False):
    r"""
    Check if parallelepiped is valid.

    Parameters
    ----------
    a : float
        Length of the :math:`v_1` vector.
    b : float
        Length of the :math:`v_2` vector.
    c : float
        Length of the :math:`v_3` vector.
    alpha : float
        Angle between vectors :math:`v_2` and :math:`v_3`. In degrees.
    beta : float
        Angle between vectors :math:`v_1` and :math:`v_3`. In degrees.
    gamma : float
        Angle between vectors :math:`v_1` and :math:`v_2`. In degrees.
    raise_error : bool, default False
        Whether to raise error if parameters could not form a parallelepiped.

    Returns
    -------
    result: bool
        Whether the parameters could from a parallelepiped.

    Raises
    ------
    ValueError
        If parameters could not form a parallelepiped.
        Only raised if ``raise_error`` is ``True`` (it is ``False`` by default).
    """

    result = (
        compare_numerically(a, ">", 0.0, ABS_TOL)
        and compare_numerically(b, ">", 0.0, ABS_TOL)
        and compare_numerically(c, ">", 0.0, ABS_TOL)
        and compare_numerically(alpha, "<", 180.0, ABS_TOL_ANGLE)
        and compare_numerically(beta, "<", 180.0, ABS_TOL_ANGLE)
        and compare_numerically(gamma, "<", 180.0, ABS_TOL_ANGLE)
        and compare_numerically(alpha, ">", 0.0, ABS_TOL_ANGLE)
        and compare_numerically(beta, ">", 0.0, ABS_TOL_ANGLE)
        and compare_numerically(gamma, ">", 0.0, ABS_TOL_ANGLE)
        and compare_numerically(gamma, "<", alpha + beta, ABS_TOL_ANGLE)
        and compare_numerically(alpha + beta, "<", 360.0 - gamma, ABS_TOL_ANGLE)
        and compare_numerically(beta, "<", alpha + gamma, ABS_TOL_ANGLE)
        and compare_numerically(alpha + gamma, "<", 360.0 - beta, ABS_TOL_ANGLE)
        and compare_numerically(alpha, "<", beta + gamma, ABS_TOL_ANGLE)
        and compare_numerically(beta + gamma, "<", 360.0 - alpha, ABS_TOL_ANGLE)
    )

    if not result and raise_error:
        message = "Parameters could not form a parallelepiped:\n"
        message += f"a = {a}"
        if not compare_numerically(a, ">", 0.0, ABS_TOL):
            message += f" (a <= 0 with numerical tolerance: {ABS_TOL})"
        message += "\n"
        message += f"b = {b}"
        if not compare_numerically(b, ">", 0.0, ABS_TOL):
            message += f" (b <= 0 with numerical tolerance: {ABS_TOL})"
        message += "\n"
        message += f"c = {c}"
        if not compare_numerically(c, ">", 0.0, ABS_TOL):
            message += f" (c <= 0 with numerical tolerance: {ABS_TOL})"
        message += "\n"
        message += f"alpha = {alpha}\n"
        if not compare_numerically(alpha, "<", 180.0, ABS_TOL_ANGLE):
            message += f"  (alpha >= 180 with numerical tolerance: {ABS_TOL_ANGLE})\n"
        if not compare_numerically(alpha, ">", 0.0, ABS_TOL_ANGLE):
            message += f"  (alpha <= 0 with numerical tolerance: {ABS_TOL_ANGLE})\n"
        message += f"beta = {beta}\n"
        if not compare_numerically(beta, "<", 180.0, ABS_TOL_ANGLE):
            message += f"  (beta >= 180 with numerical tolerance: {ABS_TOL_ANGLE})\n"
        if not compare_numerically(beta, ">", 0.0, ABS_TOL_ANGLE):
            message += f"  (beta <= 0 with numerical tolerance: {ABS_TOL_ANGLE})\n"
        message += f"gamma = {gamma}\n"
        if not compare_numerically(gamma, "<", 180.0, ABS_TOL_ANGLE):
            message += f"  (gamma >= 180 with numerical tolerance: {ABS_TOL_ANGLE})\n"
        if not compare_numerically(gamma, ">", 0.0, ABS_TOL_ANGLE):
            message += f"  (gamma <= 0 with numerical tolerance: {ABS_TOL_ANGLE})\n"
        if not compare_numerically(gamma, "<", alpha + beta, ABS_TOL_ANGLE):
            message += f"Inequality gamma < alpha + beta is not satisfied with numerical tolerance: {ABS_TOL_ANGLE}\n"
        if not compare_numerically(alpha + beta, "<", 360.0 - gamma, ABS_TOL_ANGLE):
            message += f"Inequality alpha + beta < 360 - gamma is not satisfied with numerical tolerance: {ABS_TOL_ANGLE}\n"
        if not compare_numerically(beta, "<", alpha + gamma, ABS_TOL_ANGLE):
            message += f"Inequality beta < alpha + gamma is not satisfied with numerical tolerance: {ABS_TOL_ANGLE}\n"
        if not compare_numerically(alpha + gamma, "<", 360.0 - beta, ABS_TOL_ANGLE):
            message += f"Inequality alpha + gamma < 360 - beta is not satisfied with numerical tolerance: {ABS_TOL_ANGLE}\n"
        if not compare_numerically(alpha, "<", beta + gamma, ABS_TOL_ANGLE):
            message += f"Inequality alpha < beta + gamma is not satisfied with numerical tolerance: {ABS_TOL_ANGLE}\n"
        if not compare_numerically(beta + gamma, "<", 360.0 - alpha, ABS_TOL_ANGLE):
            message += f"Inequality beta + gamma < 360 - alpha is not satisfied with numerical tolerance: {ABS_TOL_ANGLE}\n"
        raise ValueError(message)

    return result


def absolute_to_relative(basis, vector):
    r"""
    Compute relative coordinates of the vector with respect to the basis.

    Parameters
    ----------
    basis : (3, 3) |array_like|_
        Lattice vectors.
    vector : (3,) |array_like|_
        Absolute coordinates.

    Returns
    -------
    relative : (3,) :numpy:`ndarray`
        Relative coordinates.
    """

    # Three vectors of the cell
    v1 = np.array(basis[0], dtype=float)
    v2 = np.array(basis[1], dtype=float)
    v3 = np.array(basis[2], dtype=float)

    v = np.array(vector, dtype=float)
    if (v == np.zeros(3)).all():
        return np.zeros(3)

    # Compose system of linear equations
    B = np.array([np.dot(v1, v), np.dot(v2, v), np.dot(v3, v)])
    A = np.array(
        [
            [np.dot(v1, v1), np.dot(v1, v2), np.dot(v1, v3)],
            [np.dot(v2, v1), np.dot(v2, v2), np.dot(v2, v3)],
            [np.dot(v3, v1), np.dot(v3, v2), np.dot(v3, v3)],
        ]
    )

    # Solve and return
    return np.linalg.solve(A, B)


def span_orthonormal_set(vec):
    r"""
    Span orthonormal set of vectors.

    Parameters
    ----------
    vec : (3,) |array_like|_
        Vector, which serves as :math:`e_3`

    Returns
    -------
    e1 : (3,) :numpy:`ndarray`
    e2 : (3,) :numpy:`ndarray`
    e3 : (3,) :numpy:`ndarray`
    """

    vec = np.array(vec) / np.linalg.norm(vec)

    if np.allclose(vec, [0, 0, 1]):
        return (
            np.array([1.0, 0.0, 0.0]),
            np.array([0.0, 1.0, 0.0]),
            np.array([0.0, 0.0, 1.0]),
        )

    if np.allclose(vec, [0, 0, -1]):
        return (
            np.array([1.0, 0.0, 0.0]),
            np.array([0.0, -1.0, 0.0]),
            np.array([0.0, 0.0, -1.0]),
        )

    z_dir = [0, 0, 1]
    n = (
        np.cross(vec, z_dir)
        / np.linalg.norm(np.cross(vec, z_dir))
        * np.arccos(np.dot(vec, z_dir) / np.linalg.norm(vec))
    )
    rotation_matrix = Rotation.from_rotvec(n).as_matrix()

    return rotation_matrix



# RAD-tools - program for spin Hamiltonian and magnons.
# Copyright (C) 2022-2024  Andrey Rybakov
#
# e-mail: anry@uv.es, web: adrybakov.com
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

r"""
Input-output for |Vampire|_.
"""

__all__ = ["dump_vampire", "dump_mat", "dump_ucf"]

import numpy as np

from radtools.decorate.array import print_2d_array
from radtools.decorate.stats import logo
from radtools.spinham.parameter import ExchangeParameter

meV_TO_J = 1.602176634e-22


def dump_vampire(
    spinham,
    seedname="vampire",
    anisotropic=True,
    dmi=True,
    custom_mask=None,
    decimals=5,
    materials=None,
    nologo=False,
):
    """
    Save the Hamiltonian in a Human-readable format.

    Parameters
    ----------
    spinham : :py:class:`.SpinHamiltonian`
        Spin Hamiltonian to be saved.
    seedname : str, default "vampire"
        Seedname for the .UCF and .mat files. Extensions are added automatically.
        Input file always have the name "input".
    anisotropic : bool, default True
        Whether to output anisotropic exchange.
    dmi : bool, default True
        Whether to output DMI exchange.
    custom_mask : func, optional
        Custom mask for the exchange parameter. Function which take (3,3) numpy:`ndarray`
        as an input and returns (3,3) numpy:`ndarray` as an output. If given, then
        ``anisotropic`` and ``dmi`` parameters are ignored.
    decimals : int, default 4
        Number of decimals to be printed (only for the exchange values).
    materials : list of int, optional
        List of materials for the atoms. Has to have the same length as the number of
        magnetic atoms in the ``spinham``. Order is the same as in :py:attr:`.SpinHamiltonian.magnetic_atoms`.
        If not given, each atom will be considered as a separate material. Starting from 0.
    nologo : bool, default False
        Whether to print the logo in the output files.

    Returns
    -------
    content : str
        Content of the .UCF file if ``filename`` is not given.
    """

    dump_ucf(
        spinham,
        filename=seedname + ".UCF",
        anisotropic=anisotropic,
        dmi=dmi,
        custom_mask=custom_mask,
        decimals=decimals,
        materials=materials,
        nologo=nologo,
    )
    dump_mat(
        spinham,
        filename=seedname + ".mat",
        materials=materials,
        nologo=nologo,
    )
    with open("input", "w", encoding="utf-8") as file:
        file.write(
            "#------------------------------------------\n"
            f"material:file={seedname}.mat\n"
            f"material:unit-cell-file = {seedname}.UCF\n"
            "#------------------------------------------\n"
        )


def dump_mat(spinham, filename=None, materials=None, nologo=False):
    """
    Write .mat file for |Vampire|_.

    Parameters
    ----------
    spinham : :py:class:`.SpinHamiltonian`
        Spin Hamiltonian to be saved.
    filename : str, optional
        Name for the .mat file. No extensions is added automatically.
        If not given, the output is returned as a string.
    materials : list of int, optional
        List of materials for the atoms. Has to have the same length as the number of
        magnetic atoms in the ``spinham``. Order is the same as in :py:attr:`.SpinHamiltonian.magnetic_atoms`.
        If not given, each atom will be considered as a separate material. Starting from 0.
    nologo : bool, default False
        Whether to print the logo in the output files.
    """
    if nologo:
        result = []
    else:
        result = [logo(comment=True, date_time=True) + "\n"]
    if materials is not None:
        result.append(f"material:num-materials = {max(materials)+1}\n")
    else:
        result.append(f"material:num-materials = {len(spinham.magnetic_atoms)}\n")
    for i, atom in enumerate(spinham.magnetic_atoms):
        if materials is None or materials[i] not in materials[:i]:
            if materials is not None:
                m_i = materials[i] + 1
            else:
                m_i = i + 1
            result.append("#---------------------------------------------------\n")
            result.append(f"# Material {m_i} \n")
            result.append("#---------------------------------------------------\n")
            result.append(f"material[{m_i}]:material-name = {atom.name}\n")
            result.append(f"material[{m_i}]:material-element = {atom.type}\n")
            result.append(f"material[{m_i}]:atomic-spin-moment={atom.spin}\n")
            if atom._spin_direction is not None:
                result.append(
                    f"material[{m_i}]:initial-spin-direction = {atom.spin_direction[0]:.8f},{atom.spin_direction[1]:.8f},{atom.spin_direction[2]:.8f}\n"
                )
            else:
                result.append(f"material[{m_i}]:initial-spin-direction = random\n")
            result.append(f"material[{m_i}]:dampling-constant=0.1\n")
            result.append(f"material[{m_i}]:uniaxial-anisotropy-constant=0.0\n")
    result.append("#---------------------------------------------------\n")

    result = "".join(result)

    if filename is None:
        return "".join(result)

    with open(filename, "w", encoding="utf-8") as file:
        file.write("".join(result))


def dump_ucf(
    spinham,
    filename=None,
    anisotropic=True,
    dmi=True,
    custom_mask=None,
    decimals=5,
    materials=None,
    nologo=False,
):
    """
    Save the Hamiltonian in a Human-readable format.

    Parameters
    ----------
    spinham : :py:class:`.SpinHamiltonian`
        Spin Hamiltonian to be saved.
    filename : str, optional
        Name for the .UCF file. No extension is added automatically.
        If not given, the output is returned as a string.
    anisotropic : bool, default True
        Whether to output anisotropic exchange.
    dmi : bool, default True
        Whether to output DMI exchange.
    custom_mask : func, optional
        Custom mask for the exchange parameter. Function which take (3,3) numpy:`ndarray`
        as an input and returns (3,3) numpy:`ndarray` as an output. If given, then
        ``anisotropic`` and ``dmi`` parameters are ignored.
    decimals : int, default 4
        Number of decimals to be printed (only for the exchange values).
    materials : list of int, optional
        List of materials for the atoms. Has to have the same length as the number of
        magnetic atoms in the ``spinham``. Order is the same as in :py:attr:`.SpinHamiltonian.magnetic_atoms`.
        If not given, each atom will be considered as a separate material.
    nologo : bool, default False
        Whether to print the logo in the output files.

    Returns
    -------
    content : str
        Content of the .UCF file if ``filename`` is not given.
    """
    original_notation = spinham.notation
    spinham.notation = "vampire"
    if nologo:
        result = []
    else:
        result = [logo(comment=True, date_time=True) + "\n"]
    result.append("# Unit cell size:\n")
    result.append(f"{spinham.a:.8f} {spinham.b:.8f} {spinham.c:.8f}\n")
    result.append("# Unit cell lattice vectors:\n")
    result.append(
        print_2d_array(
            spinham.a1, borders=False, print_result=False, fmt=".8f"
        ).lstrip()
        + "\n"
    )
    result.append(
        print_2d_array(
            spinham.a2, borders=False, print_result=False, fmt=".8f"
        ).lstrip()
        + "\n"
    )
    result.append(
        print_2d_array(
            spinham.a3, borders=False, print_result=False, fmt=".8f"
        ).lstrip()
        + "\n"
    )
    result.append("# Atoms\n")
    result.append(f"{len(spinham.magnetic_atoms)} {len(np.unique(materials))}\n")
    atom_indices = {}
    for a_i, atom in enumerate(spinham.magnetic_atoms):
        result.append(
            f"{a_i} {atom.position[0]:.8f} {atom.position[1]:.8f} {atom.position[2]:.8f}"
        )
        if materials is not None:
            result.append(f" {materials[a_i]}")
        else:
            result.append(f" {a_i}")
        result.append("\n")
        atom_indices[atom] = a_i
    result.append("# Interactions\n")
    result.append(f"{len(spinham)} tensorial\n")

    IID = 0
    for atom1, atom2, (i, j, k), J in spinham:
        if custom_mask is not None:
            J = ExchangeParameter(custom_mask(J))
        else:
            if not dmi:
                J = ExchangeParameter(matrix=J.matrix - J.dmi_matrix)
            if not anisotropic:
                J = ExchangeParameter(matrix=J.matrix - J.aniso)
        J = J * meV_TO_J
        fmt = f"{7+decimals}.{decimals}e"
        result.append(
            f"{IID:<2} {atom_indices[atom1]:>2} {atom_indices[atom2]:>2} {i:>2} {j:>2} {k:>2} "
        )
        result.append(f"{J.xx:{fmt}} {J.xy:{fmt}} {J.xz:{fmt}} ")
        result.append(f"{J.yx:{fmt}} {J.yy:{fmt}} {J.yz:{fmt}} ")
        result.append(f"{J.zx:{fmt}} {J.zy:{fmt}} {J.zz:{fmt}}\n")
        IID += 1

    spinham.notation = original_notation

    result = "".join(result)

    if filename is None:
        return result

    with open(filename, "w", encoding="utf-8") as file:
        file.write(result)



# RAD-tools - program for spin Hamiltonian and magnons.
# Copyright (C) 2022-2024  Andrey Rybakov
#
# e-mail: anry@uv.es, web: adrybakov.com
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import os
from argparse import ArgumentParser

from termcolor import cprint

from radtools.dos.dos import DOSQE, detect_seednames
from radtools.dos.fatbands_plotting import plot_custom_fatbands
from radtools.dos.plotting import COLOURS


def manager(
    input_folder=".",
    seedname=None,
    output_name="",
    energy_window=None,
    k_window=None,
    efermi=0.0,
    verbose=False,
    interactive=False,
    separate=False,
    save_pickle=False,
    save_txt=False,
    custom=None,
    colours=None,
    labels=None,
    legend_fontsize=12,
    axes_labels_fontsize=14,
    title_fontsize=18,
    k_points=None,
):
    r"""
    :ref:`rad-plot-fatbands` script.

    Full documentation on the behaviour is available in the
    :ref:`User Guide <rad-plot-fatbands>`.
    Parameters of the function directly
    correspond to the arguments of the script.

    Parameters
    ----------
    input_folder : str, default "."
        Relative or absolute path to the folder with PDOS files.

        Console argument: ``-if`` / ``--input-folder``

        Metavar: "path"
    seedname : str, optional
        Prefix for input files with PDOS(E).

        In the case of Quantum Espresso-produced seedname is the same
        as specified in the QE projwfc.x input file (filpdos).

        If it is not provided the script tries to
        detect it automatically in the
        ``rad-plot-fatbands_input-folder`` folder.

        Console argument: ``-s`` / ``--seedname``

        Metavar: "name"
    output_name : str, optional
        Relative or absolute path to the folder for saving outputs.

        Console argument: ``-on`` / ``--output-name``

        Metavar: "path"
    energy_window : tuple of 2 float, optional
        Energy window for the plots.

        By default the whole energy range present in the files is plotted.

        Console argument: ``-ew`` / ``--energy-window``

        Metavar: ("min", "max")
    k_window : tuple of 2 float, optional
        K-point window for the plots.

        By default the whole states/eV range is plotted.

        Console argument: ``-kw`` / ``--k-window``

        Metavar: ("min", "max")
    efermi : float, default 0.0
        Fermi energy.

        Zero is shifted to Fermi energy.

        Console argument: ``-ef`` / ``--efermi``

        Metavar: "energy"
    verbose : bool, default False
        Verbose output, propagates to the called methods.

        Console argument: ``-v`` / ``--verbose``
    interactive : bool, default False
        Interactive plotting.

        Console argument: ``-i`` / ``--interactive``
    separate : bool, default False
        Whether to plot projected DOS for each atom of the same type separately.

        Console argument: ``-sep`` / ``--separate``
    save_pickle : bool, default False
        Whether to save figures as .pickle files.

        Console argument: ``-sp`` / ``--save-pickle``
    save_txt : bool, default False
        Whether to save some data as txt files.

        Console argument: ``-st`` / ``--save-txt``
    custom : list of str, optional
        Custom PDOS plot.

        Console argument: ``--custom``

        Metavar: "description"
    colours : list of str, optional
        Colours for the relative and custom plots.

        Values are passed directly to the matplotlib as strings,
        therefore any valid value is allowed. Examples: "red" or "#FF0000".
        When ``custom`` is used the order of colours is the same as for
        the values of the ``custom``.

        Console argument: ``-cls`` / ``--colours``
    labels : list of str, optional
        Labels for the custom plots.

        Amount of labels have to be the same as the amount of ``custom`` strings, or one more.
        If one more, then first one is interpreted as the label for the background
        (Use "None" to switch it off). If the amount of argument is one more  and the first one is None,
        then the label for the total PDOS is switched off and the total PDOS itself is not plotted.

        Console argument: ``-lbs`` / ``--labels``
    legend_fontsize : int, default 12
        Fontsize of the legend.

        Console argument: ``-lfs`` / ``--legend-fontsize``

        Metavar: "fontsize"
    axes_labels_fontsize : int, default 14
        Fontsize of the labes of the axes.

        Console argument: ``-alfs`` / ``--axes-labels-fontsize``

        Metavar: "fontsize"
    title_fontsize : int, default 18
        Fontsize of the title.

        Console argument: ``-tfs`` / ``--title-fontsize``

        Metavar: "fontsize"
    k_points : list of str, optional
        Plot coordinates of high symmetry points.

        Console argument: ``-kp`` / ``--k-points``

        Metavar: "G 0.23 K 0.64 Y 1.2 ..."
    """

    out_head, out_tail = os.path.split(output_name)
    if len(out_head) == 0:
        out_head = input_folder

    if colours is None:
        colours = COLOURS

    # Detect seednames if not provided.
    if seedname is None:
        seednames = detect_seednames(input_folder)
        print(f"Following DOS seednames (filpdos) are detected:")
        for item in seednames:
            cprint(f"   * {item}", "green", attrs=["bold"])
    else:
        seednames = [seedname]

    # Work with each seedname.
    for s_i, seedname in enumerate(seednames):
        cprint(
            f"({s_i + 1}/{len(seednames)}) Start to work with {seedname} seedname",
            "yellow",
            attrs=["bold"],
        )
        # Preparations
        output_root = os.path.join(out_head, f"{out_tail}{seedname}")
        if output_root != "":
            os.makedirs(output_root, exist_ok=True)

        # Load DOS data.
        dos = DOSQE(seedname, input_folder, energy_window=energy_window, efermi=efermi)
        print(f"{dos.casename} case detected.")
        for atom in dos.atoms:
            print(f"  {len(dos.atom_numbers(atom))} {atom} detected")

        plot_custom_fatbands(
            dos=dos,
            custom=custom,
            output_root=output_root,
            energy_window=energy_window,
            k_window=k_window,
            efermi=efermi,
            interactive=interactive,
            separate=separate,
            save_pickle=save_pickle,
            save_txt=save_txt,
            colours=colours,
            labels=labels,
            legend_fontsize=legend_fontsize,
            axes_labels_fontsize=axes_labels_fontsize,
            title_fontsize=title_fontsize,
            k_points=k_points,
        )


def create_parser():
    parser = ArgumentParser()
    parser.add_argument(
        "-if",
        "--input-folder",
        default=".",
        metavar="path",
        type=str,
        help="Relative or absolute path to the folder with PDOS files.",
    )
    parser.add_argument(
        "-s",
        "--seedname",
        default=None,
        metavar="name",
        type=str,
        help="Prefix for input files with PDOS(E).",
    )
    parser.add_argument(
        "-on",
        "--output-name",
        default="",
        metavar="path",
        type=str,
        help="Relative or absolute path to the folder for saving outputs.",
    )
    parser.add_argument(
        "-ew",
        "--energy-window",
        default=None,
        metavar=("min", "max"),
        type=float,
        nargs=2,
        help="Energy window for the plots.",
    )
    parser.add_argument(
        "-kw",
        "--k-window",
        default=None,
        metavar=("min", "max"),
        type=float,
        nargs=2,
        help="K-point window for the plots.",
    )
    parser.add_argument(
        "-ef",
        "--efermi",
        default=0.0,
        metavar="energy",
        type=float,
        help="Fermi energy.",
    )
    parser.add_argument(
        "-v",
        "--verbose",
        default=False,
        action="store_true",
        help="Verbose output, propagates to the called methods.",
    )
    parser.add_argument(
        "-i",
        "--interactive",
        default=False,
        action="store_true",
        help="Interactive plotting.",
    )
    parser.add_argument(
        "-sep",
        "--separate",
        default=False,
        action="store_true",
        help="Whether to plot projected DOS for each atom of the same type separately.",
    )
    parser.add_argument(
        "-sp",
        "--save-pickle",
        default=False,
        action="store_true",
        help="Whether to save figures as .pickle files.",
    )
    parser.add_argument(
        "-st",
        "--save-txt",
        default=False,
        action="store_true",
        help="Whether to save some data as txt files.",
    )
    parser.add_argument(
        "--custom",
        default=None,
        metavar="description",
        type=str,
        nargs="*",
        help="Custom PDOS plot.",
    )
    parser.add_argument(
        "-cls",
        "--colours",
        default=None,
        type=str,
        nargs="*",
        help="Colours for the relative and custom plots.",
    )
    parser.add_argument(
        "-lbs",
        "--labels",
        default=None,
        type=str,
        nargs="*",
        help="Labels for the custom plots.",
    )
    parser.add_argument(
        "-lfs",
        "--legend-fontsize",
        default=12,
        metavar="fontsize",
        type=int,
        help="Fontsize of the legend.",
    )
    parser.add_argument(
        "-alfs",
        "--axes-labels-fontsize",
        default=14,
        metavar="fontsize",
        type=int,
        help="Fontsize of the labes of the axes.",
    )
    parser.add_argument(
        "-tfs",
        "--title-fontsize",
        default=18,
        metavar="fontsize",
        type=int,
        help="Fontsize of the title.",
    )
    parser.add_argument(
        "-kp",
        "--k-points",
        default=None,
        metavar="G 0.23 K 0.64 Y 1.2 ...",
        type=str,
        nargs="*",
        help="Plot coordinates of high symmetry points.",
    )

    return parser



# RAD-tools - program for spin Hamiltonian and magnons.
# Copyright (C) 2022-2024  Andrey Rybakov
#
# e-mail: anry@uv.es, web: adrybakov.com
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from math import pi

import numpy as np
from tqdm import tqdm

__all__ = ["dipole_dipole_energy", "dipole_dipole_interaction"]

CONSTANT = 1.25663706212 * 9.2740100783**2 * 6.241509074 / 1000 / 4 / pi


def dipole_dipole_energy(magnetic_centres, progress_bar=True, normalize=True):
    r"""
    Computes magnetic dipole-dipole energy.

    This function computes the magnetic dipole-dipole energy of the set of magnetic centres:

    .. math::

        E = -\frac{\mu_0}{4\pi}\sum_{i > j}\frac{1}{\vert r_{ij}\vert^3}\left(3(\vec{m_i} \cdot \vec{r_{ij}})(\vec{m_j} \cdot \vec{r_{ij}}) - (\vec{m_i}\cdot\vec{m_j})\right)

    Parameters
    ----------
    magnetic_centres: (N, 2, 3) |array_like|_
        List of N magnetic centres.
        First element along second axis is magnetic moment (in Bohr magnetons).
        Second element along second axis if position (in Angstrom).
    progress_bar : bool, default True
        Whether to show progressbar.
    normalize : bool, default True
        Whether to normalize energy to the number of magnetic centres.

    Returns
    -------
    energy : float
        Dipole-dipole energy of the system of ``magnetic_centres``.
        Normalized to the number of magnetic centres N.

    See Also
    --------
    dipole_dipole_interaction : between two sets of magnetic centres.
    """

    magnetic_centres = np.array(magnetic_centres)

    energy = 0
    if progress_bar:
        for_range = tqdm(range(len(magnetic_centres) - 1))
    else:
        for_range = range(len(magnetic_centres) - 1)

    N = len(magnetic_centres)

    for i in for_range:
        m_i = magnetic_centres[i, 0]
        r_i = magnetic_centres[i, 1]
        r_ij = magnetic_centres[i + 1 :, 1] - r_i
        r_ij_norm = np.linalg.norm(r_ij, axis=1)
        r_ij_norm_5 = (r_ij.T / r_ij_norm**5).T
        first_term = np.sum(
            r_ij_norm_5.T * np.diag(magnetic_centres[i + 1 :, 0] @ r_ij.T),
            axis=1,
        )
        second_term = np.sum((magnetic_centres[i + 1 :, 0].T) / r_ij_norm**3, axis=1)
        energy += m_i @ (-3 * first_term + second_term)

    if normalize:
        energy /= N

    return energy * CONSTANT


def dipole_dipole_interaction(
    magnetic_centres1, magnetic_centres2, progress_bar=True, normalize=True
):
    r"""
    Computes magnetic dipole-dipole interaction.

    This function computes dipole-dipole interaction between two sets of magnetic
    centres (:math:`mc_1` and :math:`mc_2`):

    .. math::

        E = -\frac{\mu_0}{4\pi}\sum_{i \in mc_1, j \in mc_2}\frac{1}{\vert r_{ij}\vert^3}\left(3(\vec{m_i} \cdot \vec{r_{ij}})(\vec{m_j} \cdot \vec{r_{ij}}) - (\vec{m_i}\cdot\vec{m_j})\right)

    Parameters
    ----------
    magnetic_centres1: (N, 2, 3) |array_like|_
        List of N magnetic centres of the first set.
        First element along second axis is magnetic moment (in Bohr magnetons).
        Second element along second axis if position (in Angstrom).
    magnetic_centres2: (M, 2, 3) |array_like|_
        List of M magnetic centres of the second set.
        First element along second axis is magnetic moment (in Bohr magnetons).
        Second element along second axis if position (in Angstrom).
        Python cycle is running along this set.
    progress_bar : bool, default True
        Whether to show progressbar.
    normalize : bool, default True
        Whether to normalize energy to the number of magnetic centres.

    Returns
    -------
    energy : float
        Magnetic dipole-dipole interaction between ``magnetic_centres1`` and ``magnetic_centres2``.
        Normalized to :math:`N\cdot M`.

    See Also
    --------
    dipole_dipole_energy : within one set of magnetic centres.
    """

    magnetic_centres1 = np.array(magnetic_centres1)
    magnetic_centres2 = np.array(magnetic_centres2)
    energy = 0
    if progress_bar:
        for_range = tqdm(range(len(magnetic_centres2) - 1))
    else:
        for_range = range(len(magnetic_centres2) - 1)

    NM = len(magnetic_centres1) + len(magnetic_centres2)

    for i in for_range:
        m_i = magnetic_centres2[i, 0]
        r_i = magnetic_centres2[i, 1]
        r_ij = magnetic_centres1[:, 1] - r_i
        r_ij_norm = np.linalg.norm(r_ij, axis=1)
        r_ij_norm_5 = (r_ij.T / r_ij_norm**5).T
        first_term = np.sum(
            r_ij_norm_5.T * np.diag(magnetic_centres1[:, 0] @ r_ij.T),
            axis=1,
        )
        second_term = np.sum((magnetic_centres1[:, 0].T) / r_ij_norm**3, axis=1)
        energy += m_i @ (-3 * first_term + second_term)

    if normalize:
        energy /= NM

    return energy * CONSTANT



# RAD-tools - program for spin Hamiltonian and magnons.
# Copyright (C) 2022-2024  Andrey Rybakov
#
# e-mail: anry@uv.es, web: adrybakov.com
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

r"""
Input-output from |TB2J|_.
"""
__all__ = ["load_tb2j_model", "read_tb2j_model"]

import numpy as np

from radtools.crystal.constants import REL_TOL
from radtools.spinham.hamiltonian import SpinHamiltonian


def load_tb2j_model(filename, quiet=True, bravais_type=None) -> SpinHamiltonian:
    r"""
    Read spin Hamiltonian from |TB2J|_ output file.

    .. versionchanged:: 0.8.1 Renamed from ``read_tb2j_model``


    In |TB2J|_ spin Hamiltonian is define in a following notation:

    .. math::
        H = -\sum_{i,j} \hat{\boldsymbol{S}}_i \cdot \boldsymbol{J}_{i,j} \hat{\boldsymbol{S}}_j

    where spin vectors :math:`\boldsymbol{S}_i` are normalized to 1
    and double counting is present (both :math:`ij` and :math:`ji` are in the sum).
    :py:class:`.SpinHamiltonian` can store exchange values in any notation.
    One could check the notation by calling the attribute :py:attr:`.SpinHamiltonian.notation`.

    This function reads and stores exchange parameters in the notation of Hamiltonian
    mentioned above.

    Distance between the atoms are not read from the atoms,
    but rather computed from the unit cell and atom positions
    (see :py:meth:`.Crystal.get_distance`). If the distance read from the file
    is different from the computed one and ``quiet=False``, the warning is printed.
    It is usual for the computed and read distances to differ in the last digits, since
    the unit cell is provided with the same precision as the distance in
    |TB2J|_ output file.

    Parameters
    ----------
    filename : str
        Path to the |TB2J|_ output file.
    quiet : bool, default True
        Whether to suppress output.
    bravais_type : str, default None
        Expected bravais lattice type. If ``None``, the lattice type is identified
        automatically. See :py:meth:`.Crystal.identify` for more details.
        The bravais lattice type is reached by reducing the accuracy for the :ref:`library_lepage`.
        If the desired lattice type is not reached, the error is raised.

    Returns
    -------
    model : :py:class:`.SpinHamiltonian`
        Spin Hamiltonian build from |TB2J|_ file. With notation set to "TB2J".

    Raises
    ------
    ValueError
        If ``bravais_type`` is provided and can not be reached.
    """

    major_sep = "=" * 90
    minor_sep = "-" * 88
    garbage = str.maketrans(
        {"(": None, ")": None, "[": None, "]": None, ",": None, "'": None}
    )
    # Do not correct spelling, it is taken from TB2J.
    cell_flag = "Cell (Angstrom):"
    atoms_flag = "Atoms:"
    atom_end_flag = "Total"
    exchange_flag = "Exchange:"
    iso_flag = "J_iso:"
    aniso_flag = "J_ani:"
    dmi_flag = "DMI:"

    file = open(filename, "r")
    model = SpinHamiltonian(notation="TB2J")
    line = True

    # Read everything before exchange
    while line:
        line = file.readline()

        # Read cell and define eps_rel
        if line and cell_flag in line:
            a = file.readline().split()
            b = file.readline().split()
            c = file.readline().split()

            try:
                n_a = min(*tuple(map(lambda x: len(x.split(".")[1]), a)))
                n_b = min(*tuple(map(lambda x: len(x.split(".")[1]), b)))
                n_c = min(*tuple(map(lambda x: len(x.split(".")[1]), c)))
                model.eps_rel = 10 ** (-min(n_a, n_b, n_c))
            except:
                model.eps_rel = REL_TOL

            model.cell = np.array(
                [
                    list(map(float, a)),
                    list(map(float, b)),
                    list(map(float, c)),
                ]
            )

        # Read atoms
        if line and atoms_flag in line:
            line = file.readline()
            line = file.readline()
            line = file.readline().split()
            i = 1
            while line and atom_end_flag not in line:
                try:
                    # Slicing is not used intentionally.
                    magmom = tuple(map(float, [line[5], line[6], line[7]]))
                except IndexError:
                    magmom = None
                try:
                    charge = float(line[4])
                except IndexError:
                    charge = None
                model.add_atom(
                    name=line[0],
                    position=tuple(map(float, line[1:4])),
                    index=i,
                    magmom=magmom,
                    charge=charge,
                    relative=False,
                )
                line = file.readline().split()
                i += 1

        # Check if the exchange section is reached
        if line and exchange_flag in line:
            break

    # Try to reach desired bravais type, if any
    if bravais_type is not None:
        while model.eps_rel < 1 and model.type() != bravais_type:
            model.eps_rel *= 10
        if model.type() != bravais_type:
            raise ValueError(f"Bravais type {bravais_type} could not be reached.")

    while line:
        while line and minor_sep not in line:
            line = file.readline()
        line = file.readline().translate(garbage).split()
        atom1 = line[0]
        atom2 = line[1]
        R = tuple(map(int, line[2:5]))
        distance = float(line[-1])
        iso = None
        aniso = None
        dmi = None
        while line and minor_sep not in line:
            line = file.readline()

            # Read isotropic exchange
            if line and iso_flag in line:
                iso = float(line.split()[-1])

            # Read anisotropic exchange
            if line and aniso_flag in line:
                aniso = np.array(
                    [
                        list(map(float, file.readline().translate(garbage).split())),
                        list(map(float, file.readline().translate(garbage).split())),
                        list(map(float, file.readline().translate(garbage).split())),
                    ]
                )

            # Read DMI
            if line and dmi_flag in line:
                dmi = tuple(map(float, line.translate(garbage).split()[-3:]))

        # Adding info from the exchange block to the SpinHamiltonian structure
        model.add_bond(atom1, atom2, R, iso=iso, aniso=aniso, dmi=dmi)
        computed_distance = model.get_distance(atom1, atom2, R)
        if abs(computed_distance - distance) > 0.001 and not quiet:
            print(
                f"\nComputed distance is a different from the read one:\n"
                + f"  Computed: {computed_distance:.4f}\n  "
                + f"Read: {distance:.4f}\n"
            )

    return model


# For backward compatibility
read_tb2j_model = load_tb2j_model



# RAD-tools - program for spin Hamiltonian and magnons.
# Copyright (C) 2022-2024  Andrey Rybakov
#
# e-mail: anry@uv.es, web: adrybakov.com
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

r"""Atom class"""

from typing import Iterable

import numpy as np

from radtools.crystal.constants import ATOM_TYPES


class Atom:
    r"""
    Atom class.

    Notes
    -----
    "==" (and "!=") operation compare two atoms based on their names and indexes.
    If index of one atom is not define, then comparison raises ``ValueError``.
    For the check of the atom type use :py:attr:`Atom.type`.
    In most cases :py:attr:`Atom.name` = :py:attr:`Atom.type`.

    Parameters
    ----------
    name : str, default X
        Name of the atom.
    position : (3,) |array_like|_, default [0,0,0]
        Position of the atom in absolute coordinates.
    spin : float or (3,) |array_like|_, optional
        Spin or spin vector of the atom.
    magmom : (3,) |array_like|_, optional
        Magnetic moment of the atom.
    charge : float, optional
        Charge of the atom.
    index : int, optional
        Custom index of an atom, used differently in different scenarios.
        Combination of :py:attr:`.name` and :py:attr:`.index`
        is meant to be unique, when an atom belongs to some group
        (i.e. to :py:class:`.Crystal` or :py:class:`.SpinHamiltonian`).
    """

    def __init__(
        self,
        name="X",
        position=None,
        spin=None,
        magmom=None,
        charge=None,
        index=None,
    ) -> None:
        # Set name
        self._name = "X"
        self.name = name

        # Set index
        self._index = None
        if index is not None:
            self.index = index

        # Set position
        self._position = np.array([0.0, 0.0, 0.0])
        if position is not None:
            self.position = np.array(position)

        # Set magmom
        self._magmom = None
        if magmom is not None:
            self.magmom = magmom

        # Set charge
        self._charge = None
        if charge is not None:
            self.charge = charge

        # Set spin
        self._spin = None
        self._spin_direction = None
        self.spin_direction = [0, 0, 1]
        if isinstance(spin, Iterable):
            self.spin_vector = spin
        elif spin is not None:
            self.spin = spin

        # Set type placeholder
        self._type = None

    def __str__(self):
        return self.name

    def __format__(self, format_spec):
        return format(str(self), format_spec)

    # ==
    def __eq__(self, other) -> bool:
        if not isinstance(other, Atom):
            raise TypeError(
                f"TypeError: unsupported operand type(s) "
                + f"for ==: '{other.__class__.__name__}' and 'Atom'"
            )
        return self.name == other.name and self.index == other.index

    def __hash__(self):
        return hash(str(self.name) + str(self.index))

    # !=
    def __neq__(self, other):
        return not self == other

    @property
    def position(self):
        r"""
        Position of the atom.

        Returns
        -------
        position : (3,) :numpy:`ndarray`
            Position of the atom in absolute coordinates.
        """
        return self._position

    @position.setter
    def position(self, new_position):
        try:
            new_position = np.array(new_position, dtype=float)
        except:
            raise ValueError(
                f"New position is not array-like, new_position = {new_position}"
            )
        if new_position.shape != (3,):
            raise ValueError(
                f"New position has to be a 3 x 1 vector, shape: {new_position.shape}"
            )
        self._position = new_position

    @property
    def name(self):
        r"""
        Name of the atom.

        Returns
        -------
        name : str
            Name of the atom.
        """
        return self._name

    @name.setter
    def name(self, new_name):
        if new_name.startswith("__") or new_name.endswith("__"):
            raise ValueError(
                f"Name of the atom ({new_name}) is not valid. It cannot start/end with '__'."
            )
        self._name = new_name
        # Reset type
        self._type = None

    @property
    def type(self):
        r"""
        Type of an atom (i.e. Cr, Ni, ...).

        Returns
        -------
        type : str
            Type of the atom.
        """
        if self._type is None:
            self._type = "X"
            for i in ATOM_TYPES:
                if i.lower() in self._name.lower():
                    self._type = i
                    if len(i) == 2:
                        break
        return self._type

    @property
    def index(self):
        r"""
        Index of an atom, meant to be unique for some group of atoms.

        Returns
        -------
        index : int
            Index of the atom.

        Raises
        ------
        ValueError
            If index is not defined for the atom.
        """

        if self._index is None:
            raise ValueError(f"Index is not defined for the atom {self}.")
        return self._index

    @index.setter
    def index(self, new_index):
        self._index = new_index

    @property
    def spin(self):
        r"""
        Spin value of the atom.

        Independent of :py:attr:`.Atom.spin_direction`.

        Returns
        -------
        spin : float
            Spin value of the atom.

        Raises
        ------
        ValueError
            If spin is not defined for the atom.
        """

        if self._spin is None:
            raise ValueError(f"Spin value is not defined for the atom {self.fullname}.")
        return self._spin

    @spin.setter
    def spin(self, new_spin):
        self._spin = float(new_spin)

    @property
    def spin_direction(self):
        r"""
        Classical spin direction of the atom.

        .. math::

            \vec{n} = (n_x, n_y, n_z), \vert \vec{n}\vert = 1

        Returns
        -------
        spin_direction : (3,) :numpy:`ndarray`
            Classical spin direction of the atom.

        Raises
        ------
        ValueError
            If spin direction is not defined for the atom.
        """

        return self._spin_direction

    @spin_direction.setter
    def spin_direction(self, new_spin_direction):
        try:
            new_spin_direction = np.array(new_spin_direction, dtype=float)
            new_spin_direction /= np.linalg.norm(new_spin_direction)
        except BufferError:
            raise ValueError(
                f"New spin direction is not array-like, new_spin_direction = {new_spin_direction}"
            )
        if new_spin_direction.shape != (3,):
            raise ValueError(
                f"New spin direction has to be a 3 x 1 vector, shape: {new_spin_direction.shape}"
            )
        self._spin_direction = new_spin_direction

    @property
    def spin_vector(self):
        r"""
        Classical spin vector of the atom.

        .. math::

            \vec{S} = (S_x, S_y, S_z), \vert \vec{S}\vert = S

        Returns
        -------
        spin_vector : (3,) :numpy:`ndarray`
            Classical spin vector of the atom.

        Raises
        ------
        ValueError
            If :py:attr:`spin` or :py:meth:`spin_direction` is not defined for the atom.
        """

        return self.spin_direction * self.spin

    @spin_vector.setter
    def spin_vector(self, new_spin_vector):
        try:
            new_spin_vector = np.array(new_spin_vector, dtype=float)
        except:
            raise ValueError(
                f"New spin vector is not array-like, new_spin_direction = {new_spin_vector}"
            )
        if new_spin_vector.shape != (3,):
            raise ValueError(
                f"New spin vector has to be a 3 x 1 vector, shape: {new_spin_vector.shape}"
            )
        self._spin_direction = new_spin_vector / np.linalg.norm(new_spin_vector)
        self._spin = np.linalg.norm(new_spin_vector)

    @property
    def magmom(self):
        r"""
        Magnetic moment of the atom.

        Implementation is fully independent of the atom`s spin.

        .. code-block:: python

            magmom = [m_x, m_y, m_z]

        units - :math:`\mu_B`

        Returns
        -------
        magmom : (3,) :numpy:`ndarray`
            Magnetic moment of the atom.
        """

        if self._magmom is None:
            raise ValueError(
                f"Magnetic moment is not defined for the atom {self.fullname}."
            )
        return self._magmom

    @magmom.setter
    def magmom(self, new_magmom):
        try:
            new_magmom = np.array(new_magmom, dtype=float)
        except:
            raise ValueError(
                f"New magnetic moment value is not array-like, new_magmom = {new_magmom}"
            )
        if new_magmom.shape != (3,):
            raise ValueError(
                f"New magnetic moment has to be a 3 x 1 vector, shape: {new_magmom.shape}"
            )
        self._magmom = new_magmom

    @property
    def charge(self):
        r"""
        Charge of the atom.

        Returns
        -------
        charge : float
            Charge of the atom.
        """

        if self._charge is None:
            raise ValueError(f"Charge is not defined for the atom {self.fullname}.")
        return self._charge

    @charge.setter
    def charge(self, new_charge):
        self._charge = float(new_charge)

    @property
    def fullname(self):
        r"""
        Fullname (name__index) of an atom.

        Double "_" is used intentionally, so the user can use "_" for
        the name of the atom.

        If index is not defined, then only name is returned.

        Returns
        -------
        fullname : str
            Fullname of the atom.

        Raises
        ------
        ValueError
            If index is not defined for the atom.
        """
        try:
            return f"{self.name}__{self.index}"
        except ValueError:
            return self.name
